#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants as n_constants

GBP_FLOW_CLASSIFIER = 'gbp_flowclassifier'
GBP_PORT = 'gbp_port'
GBP_NETWORK_VRF = 'gbp_network_vrf'
GBP_NETWORK_EPG = 'gbp_network_epg'
GBP_NETWORK_LINK = 'gbp_network_link'

DEVICE_OWNER_SNAT_PORT = 'apic:snat-pool'
DEVICE_OWNER_SVI_PORT = 'apic:svi'

IPV4_ANY_CIDR = '0.0.0.0/0'
IPV4_METADATA_CIDR = '169.254.0.0/16'

PROMISCUOUS_TYPES = [n_constants.DEVICE_OWNER_DHCP,
                     n_constants.DEVICE_OWNER_LOADBALANCER]
PROMISCUOUS_SUFFIX = 'promiscuous'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Exceptions used by Group Policy plugin and drivers."""

from neutron_lib import exceptions

from gbpservice._i18n import _


class GroupPolicyDriverError(exceptions.NeutronException):
    """Policy driver call failed."""
    message = _("%(method)s failed.")


class GroupPolicyException(exceptions.NeutronException):
    """Base for policy driver exceptions returned to user."""
    pass


class GroupPolicyDeploymentError(GroupPolicyException):
    message = _("Deployment not configured properly. See logs for details.")


class GroupPolicyInternalError(GroupPolicyException):
    message = _("Unexpected internal failure. See logs for details.")


class GroupPolicyBadRequest(exceptions.BadRequest, GroupPolicyException):
    """Base for policy driver exceptions returned to user."""
    pass


class GroupPolicyNotSupportedError(GroupPolicyBadRequest):
    message = _("Operation %(method_name)s for resource "
                "%(resource_name)s is not supported by this "
                "deployment.")


class PolicyTargetRequiresPolicyTargetGroup(GroupPolicyBadRequest):
    message = _("An policy target group was not specified when "
                "creating policy_target.")


class PolicyTargetGroupUpdateOfPolicyTargetNotSupported(GroupPolicyBadRequest):
    message = _("Updating policy target group of policy target "
                "is not supported.")


class PolicyTargetGroupSubnetRemovalNotSupported(GroupPolicyBadRequest):
    message = _("Removing a subnet from an policy target group is not "
                "supported.")


class L2PolicyUpdateOfPolicyTargetGroupNotSupported(GroupPolicyBadRequest):
    message = _("Updating L2 policy of policy target group is not supported.")


class L3PolicyUpdateOfL2PolicyNotSupported(GroupPolicyBadRequest):
    message = _("Updating L3 policy of L2 policy is not supported.")


class UnsettingInjectDefaultRouteOfL2PolicyNotSupported(GroupPolicyBadRequest):
    message = _("Unsetting inject_default_route attribute of L2 policy is not "
                "supported.")


class L3PolicyMultipleRoutersNotSupported(GroupPolicyBadRequest):
    message = _("L3 policy does not support multiple routers.")


class L3PolicyRoutersUpdateNotSupported(GroupPolicyBadRequest):
    message = _("Updating L3 policy's routers is not supported.")


class NoSubnetAvailable(exceptions.ResourceExhausted, GroupPolicyException):
    message = _("No subnet is available from l3 policy's pool.")


class PolicyTargetGroupInUse(GroupPolicyBadRequest):
    message = _("Policy Target Group %(policy_target_group)s is in use")


class InvalidPortForPTG(GroupPolicyBadRequest):
    message = _("Subnet %(port_subnet_id)s of port %(port_id)s does not "
                "match subnet %(ptg_subnet_id)s of Policy Target Group "
                "%(policy_target_group_id)s.")


class InvalidPortExtraAttributes(GroupPolicyBadRequest):
    message = _("Port extra attribute %(attribute)s is invalid for the "
                "following reason: %(reason)s")


class InvalidSubnetForPTG(GroupPolicyBadRequest):
    message = _("Subnet %(subnet_id)s does not belong to network "
                "%(network_id)s associated with L2P %(l2p_id)s for PTG "
                "%(ptg_id)s.")


class OverlappingIPPoolsInSameTenantNotAllowed(GroupPolicyBadRequest):
    message = _("IP Pool %(ip_pool)s overlaps with one of the existing L3P "
                "for the same tenant %(overlapping_pools)s.")


class SharedResourceReferenceError(GroupPolicyBadRequest):
    message = _("Shared resource of type %(res_type)s with id %(res_id)s "
                "can't reference the non shared resource of type "
                "%(ref_type)s with id %(ref_id)s")


class InvalidSharedResource(GroupPolicyBadRequest):
    message = _("Resource of type %(type)s cannot be shared by driver "
                "%(driver)s")


class CrossTenantL2PolicyL3PolicyNotSupported(GroupPolicyBadRequest):
    message = _("Cross tenancy not supported between L2Ps and L3Ps")


class CrossTenantPolicyTargetGroupL2PolicyNotSupported(
        GroupPolicyBadRequest):
    message = _("Cross tenancy not supported between PTGs and L2Ps")


class NonSharedNetworkOnSharedL2PolicyNotSupported(GroupPolicyBadRequest):
    message = _("Non Shared Network can't be set for a shared L2 Policy")


class InvalidSharedAttributeUpdate(GroupPolicyBadRequest):
    message = _("Invalid shared attribute update. Shared resource %(id)s is "
                "referenced by %(rid)s, which is either shared or owned by a "
                "different tenant.")


class ExternalRouteOverlapsWithL3PIpPool(GroupPolicyBadRequest):
    message = _("Destination %(destination)s for ES %(es_id)s overlaps with "
                "L3P %(l3p_id)s.")


class ExternalSegmentSubnetOverlapsWithL3PIpPool(GroupPolicyBadRequest):
    message = _("Subnet %(subnet)s for ES %(es_id)s overlaps with "
                "L3P %(l3p_id)s.")


class ExternalRouteNextHopNotInExternalSegment(GroupPolicyBadRequest):
    message = _("One or more external routes' nexthop are not part of "
                "subnet %(cidr)s.")


class InvalidL3PExternalIPAddress(GroupPolicyBadRequest):
    message = _("Address %(ip)s allocated for l3p %(l3p_id)s on segment "
                "%(es_id)s doesn't belong to the segment subnet %(es_cidr)s")


class InvalidAttributeUpdateForES(GroupPolicyBadRequest):
    message = _("Attribute %(attribute)s cannot be updated for External "
                "Segment.")


class MultipleESPerEPNotSupported(GroupPolicyBadRequest):
    message = _("Multiple External Segments per External Policy is not "
                "supported.")


class ESIdRequiredWhenCreatingEP(GroupPolicyBadRequest):
    message = _("External Segment ID is required when creating ExternalPolicy")


class ESUpdateNotSupportedForEP(GroupPolicyBadRequest):
    message = _("external_segments update for External Policy is not "
                "supported.")


class MultipleESPerL3PolicyNotSupported(GroupPolicyBadRequest):
    message = _("Only one External Segment per L3 Policy supported.")


class InvalidSubnetForES(GroupPolicyBadRequest):
    message = _("External Segment subnet %(sub_id)s is not part of an "
                "external network %(net_id)s.")


class OnlyOneEPPerTenantAllowed(GroupPolicyBadRequest):
    message = _("Only one External Policy per Tenant is allowed.")


class ImplicitSubnetNotSupported(GroupPolicyBadRequest):
    message = _("This policy driver doesn't support implicit external "
                "subnet creation.")


class DefaultL3PolicyAlreadyExists(GroupPolicyBadRequest):
    message = _("Default L3 Policy with name %(l3p_name)s already "
                "exists and is visible for this tenant.")


class DefaultExternalSegmentAlreadyExists(GroupPolicyBadRequest):
    message = _("Default External Segment with name %(es_name)s already "
                "exists and is visible for this tenant.")


class InvalidCrossTenantReference(GroupPolicyBadRequest):
    message = _("Not supported cross tenant reference: object "
                "%(res_type)s:%(res_id)s can't link %(ref_type)s:%(ref_id)s "
                "unless it's shared.")


class InvalidNetworkAccess(GroupPolicyBadRequest):
    message = _("%(msg)s : Network id %(network_id)s doesn't belong to "
                " the tenant id %(tenant_id)s.")


class InvalidRouterAccess(GroupPolicyBadRequest):
    message = _("%(msg)s : Router id %(router_id)s does not belong to the "
                " tenant id %(tenant_id)s.")


class MultipleRedirectActionsNotSupportedForRule(GroupPolicyBadRequest):
    message = _("Resource Mapping Driver does not support multiple redirect "
                "actions in a Policy Rule.")


class MultipleRedirectActionsNotSupportedForPRS(GroupPolicyBadRequest):
    message = _("Resource Mapping Driver does not support multiple redirect "
                "actions in a Policy Rule Set.")


class InvalidNetworkServiceParameters(GroupPolicyBadRequest):
    message = _("Resource Mapping Driver currently supports Network Service "
                "Parameters with the following types and values: "
                "type: 'ip_single' value: 'self_subnet'; "
                "type: 'ip_pool' value: 'nat_pool'; "
                "type: 'ip_single' value: 'nat_pool'; "
                "type: 'qos_maxrate' value: numerical of intended Kbps; "
                "type: 'qos_burstrate' value: numerical of intended Kbps.")


class ESSubnetRequiredForNatPool(GroupPolicyBadRequest):
    message = _("Resource Mapping Driver requires an External Segment which "
                "has an external subnet specified to create a Nat Pool")


class InvalidESSubnetCidrForNatPool(GroupPolicyBadRequest):
    message = _("Resource Mapping Driver requires an External Segment which "
                "maps to ip pool value specified in the nat pool")


class NSPRequiresES(GroupPolicyBadRequest):
    message = _("Driver requires an External Segment in "
                "l3policy to associate a NSP with value nat_pool to a PTG")


class NSPRequiresNatPool(GroupPolicyBadRequest):
    message = _("Resource Mapping Driver requires an External Segment in "
                "l3policy which has nat_pool associated for associating a NSP "
                "with value nat_pool to a PTG")


class L3PEsinUseByNSP(exceptions.InUse, GroupPolicyException):
    message = _("The External Segment in L3Policy cannot be updated because "
                "it is in use by Network Service Policy")


class NatPoolinUseByNSP(exceptions.InUse, GroupPolicyException):
    message = _("The Nat Pool is in use by Network Service Policy")


class OverlappingNATPoolInES(GroupPolicyBadRequest):
    message = _("One or more NAT Pools associated with ES %(es_id)s overlaps "
                "with NAT Pool %(np_id)s.")


class OverlappingSubnetForNATPoolInES(GroupPolicyBadRequest):
    message = _("One or more subnets associated with network %(net_id)s "
                "partially overlaps with NAT Pool %(np_id)s.")


class InvalidProxiedGroupL3P(GroupPolicyBadRequest):
    message = _("Cannot proxy PTG %(ptg_id)s: it's on a different L3 policy "
                "%(l3p_id)s")


class InvalidProxiedGroupL2P(GroupPolicyBadRequest):
    message = _("Cannot proxy PTG %(ptg_id)s: it's on the same L2 Policy as "
                "the proxy group of type L2.")


class OnlyOneProxyGatewayAllowed(GroupPolicyBadRequest):
    message = _("Another proxy gateway PT already exists for group "
                "%(group_id)s")


class OnlyOneGroupDefaultGatewayAllowed(GroupPolicyBadRequest):
    message = _("Another group default gateway PT already exists for group "
                "%(group_id)s")


class PTGAlreadyProvidingRedirectPRS(GroupPolicyBadRequest):
    message = _("PTG %(ptg_id)s is already providing a redirect PRS.")


class InvalidClusterId(GroupPolicyBadRequest):
    message = _("In RMD and derived drivers, a PT cluster_id should point to "
                "an existing PT.")


class PolicyTargetInUse(GroupPolicyBadRequest):
    message = _("Cannot delete a PT in use by a cluster.")


class InvalidClusterPtg(GroupPolicyBadRequest):
    message = _("Inter PTG clustering disallowed.")


class NatPoolInUseByPort(exceptions.InUse, GroupPolicyException):
    message = _("Ports or floating IP addresses are using the subnet "
                "corresponding to Nat Pool.")


class IdenticalExternalRoute(GroupPolicyBadRequest):
    message = _("External segments %(es1)s and %(es2)s cannot have "
                "identical external route CIDR %(cidr)s if associated "
                "with a common L3 policy.")


class NoValidAddressScope(GroupPolicyBadRequest):
    message = _("No address scope was either provided, could be "
                "determined, or could be created for a l3_policy.")


class NoAddressScopeForSubnetpool(GroupPolicyBadRequest):
    message = _("Subnetpool does not have an associated address scope.")


class InconsistentAddressScopeSubnetpool(GroupPolicyBadRequest):
    message = _("Subnetpool is not associated with the address "
                "scope for a l3_policy.")


class IncorrectSubnetpoolUpdate(GroupPolicyBadRequest):
    message = _("Subnetpool %(subnetpool_id)s cannot be disassociated "
                "from L3 Policy %(l3p_id)s since it has allocated subnet(s) "
                "associated with that L3 Policy")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import uuid as pyuuid

import mock
from mock import patch
from oslo_config import cfg
import unittest2

from gbpservice.nfp.lib import transport
from gbpservice.nfp.orchestrator.db import nfp_db as nfpdb


dummy_data = {}

with mock.patch('oslo_config.cfg.CONF.register_opts') as opt:
    from gbpservice.nfp.orchestrator.modules import (
        device_orchestrator)


class DummyController(object):
    def event_complete(self, event, result=None, return_value=None):
        return

    def new_event(self, id, data, key):
        return


class DummyDesc(object):
    def to_dict(self):
        return {}


class DummyEvent(object):

    def __init__(self, data, status, ref_count=0):
        self.id = ''
        self.key = str(pyuuid.uuid4())
        self.data = {}
        self.data['status'] = status
        self.data['id'] = 'vm-id'

        self.data['network_function_id'] = 'network_function_id'
        self.data['network_function_device_id'] = 'vm-id'
        self.data['network_function_instance_id'] = (
            'network_function_instance_id')
        self.data['ports'] = [{'id': 'myid1',
                               'port_model': 'neutron',
                               'port_classification': 'management',
                               'port_role': 'active'}]
        self.data['mgmt_port_id'] = [
            {'id': 'myid1',
             'port_model': 'neutron',
             'port_classification': 'management',
             'port_role': 'active'},
        ]
        self.data['interfaces_in_use'] = 1
        self.data['reference_count'] = ref_count
        self.data['service_details'] = {'service_vendor': 'vyos'}
        self.data['network_function'] = {'id': 'network_function_id'}
        self.data['network_function_device'] = {'id': 'vm-id'}
        self.data['network_function_instance'] = {
            'id': 'network_function_instance_id'}
        self.data['resource_owner_context'] = {'admin_token': str(
            pyuuid.uuid4()), 'tenant_id': str(pyuuid.uuid4()),
            'admin_tenant_id': str(pyuuid.uuid4())}
        self.data['admin_token'] = str(pyuuid.uuid4())
        self.data['provider'] = {'ptg': None}
        self.data['consumer'] = {'ptg': None}
        self.binding_key = self.data['service_details'][
                                'service_vendor'] + self.data[
                                            'network_function']['id']
        self.context = {}
        self.desc = DummyDesc()

        self.context = self.data


class Desc(object):

    def __init__(self):
        uuid = pyuuid.uuid4()
        id = ''
        self.uuid = str(uuid) + ':' + id
        self.type = ''
        self.flag = ''
        self.worker = ''
        self.poll_desc = None

    def to_dict(self):
        return {'uuid': self.uuid,
                'type': self.type,
                'flag': self.flag,
                'worker': self.worker,
                'poll_desc': self.poll_desc
                }


class HaproxyDummyDriver(object):

    def get_network_function_device_status(self):
        pass

    def get_network_function_device_config(self, device, config,
                                           is_delete=False):
        pass


class DummyExtensionManager(object):
    drivers = 'dummy-driver'


param_req = {'param1': 'value1', 'param2': 'value2'}

cfg.CONF.import_group('keystone_authtoken', 'keystonemiddleware.auth_token')
orchestration_driver = HaproxyDummyDriver()
NDO_CLASS_PATH = ('gbpservice.nfp.orchestrator'
                  '.modules.device_orchestrator')


class NDOModuleTestCase(unittest2.TestCase):

    @mock.patch.object(device_orchestrator, 'events_init')
    @mock.patch.object(device_orchestrator, 'rpc_init')
    def test_module_init(self, mock_rpc_init, mock_events_init):
        controller = "dummy-controller"
        config = "dummy-config"
        device_orchestrator.DeviceOrchestrator = mock.Mock()
        device_orchestrator.nfp_module_init(controller, config)
        mock_events_init.assert_called_once_with(
            controller, config, device_orchestrator.DeviceOrchestrator())
        mock_rpc_init.assert_called_once_with(controller, config)

    def test_rpc_init(self):
        controller = mock.Mock()
        config = mock.Mock()
        device_orchestrator.rpc_init(controller, config)
        controller.register_rpc_agents.assert_called_once_with(mock.ANY)
        call_args, call_kwargs = controller.register_rpc_agents.call_args
        self.assertEqual(1, len(call_args[0]))
        self.assertIsInstance(call_args[0][0],
                              device_orchestrator.RpcAgent)

    def test_events_init(self):
        controller = mock.Mock()
        config = mock.Mock()
        device_orchestrator.events_init(
            controller, config, device_orchestrator)
        controller.register_events.assert_called_once_with(mock.ANY)


class NDORpcHandlerTestCase(object):

    def setUp(self):
        super(NDORpcHandlerTestCase, self).setUp()
        self.controller = mock.Mock()
        self.config = mock.Mock()
        self.rpc_handler = device_orchestrator.RpcHandler(self.config,
                                                          self.controller)

    @mock.patch.object(device_orchestrator.DeviceOrchestrator,
                       "get_network_function_config_info")
    def test_rpc_create_network_function(
            self,
            mock_get_network_function_config_info):
        self.rpc_response = {'notification_data': {
            'kwargs': [{'resource': 'healthmonitor',
                        'kwargs': {'result': 'success',
                                   'device_id': 'dev-id'}}]
        }}
        self.rpc_handler.get_network_function_config_info("context",
                                                          self.rpc_response)
        event_id = 'DEVICE_HEALTHY'
        event_data = {'device_id': 'dev-id'}
        self.controller._create_event.assert_called_once_with(
            event_id=event_id,
            event_data=event_data)
        self.rpc_handler.get_network_function_config_info.\
            assert_called_once_with("context", self.rpc_response)


@patch(NDO_CLASS_PATH + '.NDOConfiguratorRpcApi.__init__',
       mock.MagicMock(return_value=None))
class NDORpcApiTestCase(unittest2.TestCase):

    def setUp(self):
        super(NDORpcApiTestCase, self).setUp()
        self.controller = mock.Mock()
        self.config = mock.Mock()

    def test_create_network_function_device_config(self):
        context = 'context'
        conf = {'info': 'info'}
        self.rpc_handler = device_orchestrator.NDOConfiguratorRpcApi(
            context, conf)
        self.rpc_handler.conf = mock.MagicMock(return_value=conf)
        self.rpc_handler.context = mock.MagicMock(return_value=context)
        self.rpc_handler.rpc_api = mock.MagicMock(return_value=True)
        device_data =\
            {'id': 'network_function_id',
             'network_function_id': 'network_function_id',
             'network_function_instance_id': 'network_function_instance_id',
             'network_function_device_id': 'network_function_instance_id',
             'mgmt_ip_address': 'mgmt-ip',
             'service_details': {'service_type': 'service_type',
                                 'service_vendor': 'service_vendor'}}
        config_params = {'info': {'service_type': ''},
                         'config': [{'kwargs': {}}]}
        transport.send_request_to_configurator = mock.MagicMock(
            return_value=True)
        self.rpc_handler.create_network_function_device_config(device_data,
                                                               config_params)

        transport.send_request_to_configurator.assert_called_once_with(
            self.rpc_handler.conf, self.rpc_handler.context,
            config_params, 'CREATE', True)

    def test_delete_network_function_device_config(self):
        context = 'context'
        conf = 'config'
        self.rpc_handler = device_orchestrator.NDOConfiguratorRpcApi(
            context, conf)
        self.rpc_handler.conf = mock.MagicMock(return_value=conf)
        self.rpc_handler.context = mock.MagicMock(return_value=context)
        self.rpc_handler.rpc_api = mock.MagicMock(return_value=True)
        device_data =\
            {'id': 'network_function_id',
             'network_function_id': 'network_function_id',
             'network_function_instance_id': 'network_function_instance_id',
             'network_function_device_id': 'network_function_instance_id',
             'mgmt_ip_address': 'mgmt-ip',
             'service_details': {'service_type': 'service_type',
                                 'service_vendor': 'service_vendor'}}
        config_params = {'info': {'service_type': ''},
                         'config': [{'kwargs': {}}]}
        transport.send_request_to_configurator = mock.MagicMock(
            return_value=True)
        self.rpc_handler.delete_network_function_device_config(device_data,
                                                               config_params)
        transport.send_request_to_configurator.assert_called_once_with(
            self.rpc_handler.conf, self.rpc_handler.context,
            config_params, 'DELETE', True)


@patch(NDO_CLASS_PATH + '.DeviceOrchestrator._create_event',
       mock.MagicMock(return_value=True))
@patch(NDO_CLASS_PATH + '.DeviceOrchestrator.db_session',
       mock.MagicMock(return_value=True))
@patch(NDO_CLASS_PATH +
       '.DeviceOrchestrator._get_orchestration_driver',
       mock.MagicMock(return_value=orchestration_driver))
@patch(NDO_CLASS_PATH + '.NDOConfiguratorRpcApi.__init__',
       mock.MagicMock(return_value=None))
class DeviceOrchestratorTestCase(unittest2.TestCase):

    def _initialize_ndo_handler(self):
        ndo_handler = device_orchestrator.DeviceOrchestrator(
            DummyController, cfg.CONF)
        self.event = DummyEvent(dummy_data, 'PENDING_CREATE')
        return ndo_handler

    @unittest2.skip('skipping')
    @mock.patch.object(device_orchestrator.DeviceOrchestrator,
                       'device_configuration_complete')
    def test_handle_event(self, mock_device_configuration_complete):
        ndo_mgr = device_orchestrator.DeviceOrchestrator(object, cfg.CONF)
        mock_device_configuration_complete.return_value = True
        self.event = DummyEvent(dummy_data, 'DEVICE_CONFIGURED')
        self.event.id = 'DEVICE_CONFIGURED'

        ndo_mgr.handle_event(self.event)
        mock_device_configuration_complete.assert_called_with(self.event)

    @mock.patch.object(nfpdb.NFPDbBase, 'update_network_function_device')
    def test_check_device_up(self, mock_update_nfd):
        ndo_handler = self._initialize_ndo_handler()
        ndo_handler._controller = mock.MagicMock(return_value='')
        mock_update_nfd.return_value = 100
        orig_event_data = {}
        orchestration_driver.get_network_function_device_status = (
            mock.MagicMock(return_value='ACTIVE'))
        status = 'DEVICE_UP'
        orig_event_data['status'] = status
        orig_event_data['status_description'] = ndo_handler.status_map[status]
        orig_event_data['id'] = self.event.data['id']
        poll_status = ndo_handler.check_device_is_up(self.event)
        self.assertEqual(poll_status, {'poll': False})
        orchestration_driver.get_network_function_device_status = (
            mock.MagicMock(return_value='ERROR'))
        status = 'DEVICE_NOT_UP'
        orig_event_data['status'] = status
        orig_event_data['status_description'] = ndo_handler.status_map[status]
        orig_event_data['network_function_id'] = self.event.data[
            'network_function']['id']
        orig_event_data['binding_key'] = None
        orig_event_data['network_function_instance_id'] = self.event.data[
            'network_function_instance']['id']
        orig_event_data['network_function_device_id'] = self.event.data[
            'network_function_device']['id']
        poll_status = ndo_handler.check_device_is_up(self.event)
        self.assertEqual(poll_status, {'poll': False})

        mock_update_nfd.assert_called_with(ndo_handler.db_session,
                                           orig_event_data['id'],
                                           orig_event_data)
        ndo_handler._controller.reset_mock()

    @mock.patch.object(nfpdb.NFPDbBase, 'update_network_function_device')
    def test_health_check(self, mock_update_nfd):
        ndo_handler = self._initialize_ndo_handler()
        mock_update_nfd.return_value = 100

        ndo_handler.configurator_rpc.create_network_function_device_config = (
            mock.MagicMock(return_value=101))
        orchestration_driver.get_network_function_device_config = (
            mock.MagicMock(return_value=param_req))

        self.event.data['management'] = {'port': {'ip_address': '127.0.0.1'}}
        self.event.desc = Desc()
        orig_event_data = {}
        orig_event_data['id'] = self.event.data['id']
        orig_event_data['mgmt_ip_address'] = self.event.data[
            'management']['port']['ip_address']
        orig_event_data['service_details'] = self.event.data['service_details']
        orig_event_data['network_function_id'] = self.event.data[
            'network_function']['id']
        orig_event_data['network_function_instance_id'] = self.event.data[
            'network_function_instance']['id']
        event_desc = self.event.desc.to_dict()
        orig_event_data['nfp_context'] = {'event_desc': event_desc,
                                          'id': self.event.id,
                                          'key': self.event.key}
        orig_event_data['tenant_id'] = self.event.data[
                'resource_owner_context']['admin_tenant_id']
        orig_event_data['periodicity'] = 'initial'
        ndo_handler.perform_initial_health_check(self.event)
        ndo_handler.configurator_rpc.create_network_function_device_config.\
            assert_called_with(orig_event_data, param_req)

    @mock.patch.object(nfpdb.NFPDbBase, '_get_network_function_device')
    @mock.patch.object(nfpdb.NFPDbBase, 'update_network_function_device')
    def test_plug_interfaces(self, mock_update_nfd, mock_get_nfd):
        ndo_handler = self._initialize_ndo_handler()

        mock_update_nfd.return_value = 100
        orig_event_data = copy.deepcopy(self.event.data)
        ndo_handler._prepare_device_data = mock.MagicMock(
            return_value=orig_event_data)
        orig_event_data['status_description'] = ''

        orchestration_driver.plug_network_function_device_interfaces = (
            mock.MagicMock(return_value=(True)))
        ndo_handler._create_event = mock.MagicMock(return_value=True)

        orig_event_data['interfaces_in_use'] += len(orig_event_data['ports'])

        mock_get_nfd.return_value = {
            'id': 'device_id',
            'interfaces_in_use': 2
        }
        ndo_handler.plug_interfaces(self.event)
        mock_update_nfd.assert_called_with(ndo_handler.db_session,
                                           orig_event_data['id'],
                                           {'interfaces_in_use': (
                                               orig_event_data[
                                                   'interfaces_in_use'])})

        orchestration_driver.plug_network_function_device_interfaces = (
            mock.MagicMock(return_value=(False)))
        ndo_handler._create_event = mock.MagicMock(return_value=True)
        event_id = 'DEVICE_CONFIGURATION_FAILED'

        ndo_handler.plug_interfaces(self.event)
        ndo_handler._create_event.assert_called_with(
            event_id=event_id,
            event_data=orig_event_data,
            is_internal_event=True)

    @mock.patch.object(nfpdb.NFPDbBase, 'update_network_function_device')
    def test_create_device_configuration(self, mock_update_nfd):
        ndo_handler = self._initialize_ndo_handler()
        config_params = {
            'param1': 'value1',
            'parama2': 'value2',
            'config': [{'resource_data': {'forward_route': True}}]}
        orchestration_driver.get_network_function_device_config = (
            mock.MagicMock(return_value=config_params))
        ndo_handler.configurator_rpc.create_network_function_device_config = (
            mock.MagicMock(return_value=True))
        ndo_handler._create_event = mock.MagicMock(return_value=True)
        ndo_handler._controller.event_complete = mock.MagicMock(
                            return_value=None)
        self._update_event()
        device = self._get_device()
        ndo_handler.create_device_configuration(self.event)
        event_id = "UPDATE_DEVICE_CONFIG_PARAMETERS"
        nfp_context = copy.deepcopy(device['nfp_context'])
        nfp_context.update(resource_owner_context=self.event.data[
            'resource_owner_context'],
                           management=self.event.data['management'],
                           admin_token=self.event.data['admin_token'],
                           service_chain_specs=self.event.data[
                               'service_chain_specs'],
                           network_function_device_id=self.event.data[
                               'network_function_device_id'],
                           provider=self.event.data['provider'],
                           id=self.event.data['network_function_device_id'],
                           reference_count=self.event.data['reference_count'],
                           network_function_instance=self.event.data[
                               'network_function_instance'],
                           mgmt_port_id=[{'port_role': 'active',
                                          'port_model': 'neutron',
                                          'id': 'myid1',
                                          'port_classification': (
                                              'management')}],
                           network_function=self.event.data[
                               'network_function'],
                           network_function_id='network_function_id',
                           service_details=self.event.data['service_details'],
                           network_function_instance_id=(
                               'network_function_instance_id'),
                           consumer=self.event.data['consumer'],
                           ports=self.event.data['ports'],
                           interfaces_in_use=self.event.data[
                               'interfaces_in_use'],
                           status='PENDING_CREATE')
        nfp_context.pop('binding_key', None)
        nfp_context.pop('key', None)
        event_data = {'device': device, 'nfp_context': nfp_context,
                      'config_params': config_params}
        ndo_handler._create_event.assert_called_with(event_id=event_id,
                                                     event_data=event_data)

    def _update_event(self):
        self.event.data['management'] = {'port': {'ip_address': '127.0.0.1'}}
        self.event.data['provider']['port'] = {
            'ip_address': '127.0.0.1', 'mac_address': 'xx:xx:xx:xx'}
        self.event.data['consumer']['port'] = {
            'ip_address': '127.0.0.1', 'mac_address': 'xx:xx:xx:xx'}
        self.event.data['provider']['subnet'] = {
            'cidr': '11.0.0.0/24', 'gateway_ip': '11.0.0.1'}
        self.event.data['consumer']['subnet'] = {
            'cidr': '11.0.0.0/24', 'gateway_ip': '11.0.0.1'}
        self.event.data['network_function_device'][
            'mgmt_ip_address'] = self.event.data['management']['port'][
            'ip_address']
        self.event.data['service_chain_specs'] = []
        self.event.desc = Desc()

    def _get_device(self):
        device = {}
        device['mgmt_ip_address'] = self.event.data[
            'management']['port']['ip_address']
        device['consumer_gateway_ip'] = self.event.data[
            'consumer']['subnet']['gateway_ip']
        device['mgmt_ip'] = self.event.data['management']['port']['ip_address']
        device['provider_gateway_ip'] = self.event.data[
            'provider']['subnet']['gateway_ip']
        device['consumer_mac'] = self.event.data[
            'consumer']['port']['mac_address']
        device['id'] = self.event.data['id']
        device['nfp_context'] = {'event_desc': self.event.desc.to_dict(),
                                 'binding_key': self.event.binding_key,
                                 'id': self.event.id, 'key': self.event.key,
                                 'network_function_device': self.event.data[
                                     'network_function_device']}
        device['tenant_id'] = self.event.data[
            'resource_owner_context']['admin_tenant_id']
        device.update({
            'provider_mac': self.event.data['provider']['port']['mac_address'],
            'network_function_instance_id': self.event.data[
                'network_function_instance']['id'],
            'provider_ip': self.event.data['provider']['port']['ip_address'],
            'network_function_id': self.event.data['network_function']['id'],
            'service_details': self.event.data['service_details'],
            'consumer_ip': self.event.data['consumer']['port']['ip_address'],
            'consumer_cidr': self.event.data['consumer']['subnet']['cidr'],
            'provider_cidr': self.event.data['provider']['subnet']['cidr']})
        return device

    def test_update_config_params(self):
        ndo_handler = self._initialize_ndo_handler()
        ndo_handler._controller.event_complete = mock.MagicMock(
                return_value=None)
        ndo_handler.update_config_params(self.event)
        ndo_handler._create_event.assert_called_with(
                event_id='DEVICE_CONFIG_PARAMETERS_UPDATED',
                event_data=self.event.data, is_internal_event=True)

    def test_device_configuration_updated(self):
        ndo_handler = self._initialize_ndo_handler()
        ndo_handler._create_event = mock.MagicMock(return_value=True)
        ndo_handler._controller.event_complete = mock.MagicMock(
                            return_value=None)
        ndo_handler.configurator_rpc.create_network_function_device_config = (
            mock.MagicMock(return_value=True))

        config_params = {
            'param1': 'value1',
            'param2': 'value2',
            'config': [{'resource_data': {'forward_route': True}}]}
        orchestration_driver.get_network_function_device_config = (
            mock.MagicMock(return_value=config_params))
        self._update_event()
        device = self._get_device()
        self.event.data['device'] = device
        self.event.data['nfp_context'] = {'service_chain_specs': []}
        self.event.data['config_params'] = config_params
        ndo_handler.device_configuration_updated(self.event)
        ndo_handler.configurator_rpc.create_network_function_device_config.\
            assert_called_with(device, config_params)

    @mock.patch.object(nfpdb.NFPDbBase, '_get_network_function_device')
    @mock.patch.object(nfpdb.NFPDbBase, 'update_network_function_device')
    def test_device_configuration_complete(self,
                                           mock_update_nfd, mock_get_nfd):
        ndo_handler = self._initialize_ndo_handler()
        device = self.event.data
        status = 'ACTIVE'
        device = {'nfp_context': device}
        device['nfp_context']['network_function_device']['reference_count'] = 0
        device['nfp_context']['network_function_device']['status'] = status
        device['nfp_context']['network_function_device'][
            'status_description'] = ndo_handler.status_map[status]
        reference_count = device['nfp_context']['network_function_device'][
            'reference_count'] + 1
        event_desc = Desc()
        device['nfp_context']['event_desc'] = event_desc.to_dict()
        device['nfp_context']['key'] = self.event.key
        device['nfp_context']['binding_key'] = self.event.binding_key
        ndo_handler._prepare_device_data = mock.MagicMock(return_value=device)
        ndo_handler._create_event = mock.MagicMock(return_value=True)
        ndo_handler.nsf_db.get_network_function_device = (
            mock.MagicMock(return_value={'reference_count': (
                reference_count - 1)}))
        self.event.data = device
        ndo_handler._controller = mock.MagicMock(return_value=True)
        mock_get_nfd.return_value = {
            'id': 'device_id',
            'reference_count': 0
        }

        ndo_handler.device_configuration_complete(self.event)
        ndo_handler._controller.reset_mock()

    @mock.patch.object(nfpdb.NFPDbBase, 'get_network_function_device')
    @mock.patch.object(nfpdb.NFPDbBase, 'update_network_function_device')
    @mock.patch.object(nfpdb.NFPDbBase, 'get_network_function')
    @mock.patch.object(nfpdb.NFPDbBase, 'get_port_info')
    def test_delete_network_function_device(self, mock_get_port, mock_get_nf,
                                            mock_update_nfd, mock_get_nfd):
        ndo_handler = self._initialize_ndo_handler()
        delete_event_req = DummyEvent(dummy_data, 'ACTIVE')
        delete_event_req.data = \
            {'network_function_device_id': 'device-id',
             'network_function_instance': {'id': 'nfi-id',
                                           'port_info': []},
             'network_function_device': {'id': 'device-id'}}
        mgmt_port_id = {'id': 'port-id', 'port_model': 'port-policy'}
        ndo_handler._prepare_device_data_fast = mock.MagicMock(
            return_value=delete_event_req.data)
        ndo_handler._get_service_type = mock.MagicMock(
            return_value='service-type')
        ndo_handler._get_port = mock.MagicMock(return_value=mgmt_port_id)

        mock_get_port.return_value = mgmt_port_id
        mock_get_nfd.return_value = {'id': 'device-id',
                                     'mgmt_port_id': ['mgmt-data-port-id']}
        delete_event_req.data.update(
            {'event_desc': delete_event_req.desc.to_dict()})
        event_id = 'DELETE_CONFIGURATION'
        ndo_handler._create_event = mock.MagicMock(return_value=True)

        delete_event_req.context = delete_event_req.data
        ndo_handler.delete_network_function_device(delete_event_req)
        ndo_handler._create_event.assert_called_with(
            event_id=event_id,
            event_data=delete_event_req.data,
            is_internal_event=True)

    @mock.patch.object(nfpdb.NFPDbBase, 'update_network_function_device')
    def test_delete_device_configuration(self, mock_update_nfd):
        ndo_handler = self._initialize_ndo_handler()
        config_params = {'param1': 'value1', 'parama2': 'value2'}
        self.event = DummyEvent(dummy_data, 'ACTIVE')
        orchestration_driver.get_network_function_device_config = (
            mock.MagicMock(return_value=config_params))
        ndo_handler.configurator_rpc.delete_network_function_device_config = (
            mock.MagicMock(return_value=True))

        ndo_handler.delete_device_configuration(self.event)
        ndo_handler.configurator_rpc.delete_network_function_device_config.\
            assert_called_with(self.event.data, config_params)

    @mock.patch.object(nfpdb.NFPDbBase, '_get_network_function_device')
    @mock.patch.object(nfpdb.NFPDbBase, 'update_network_function_device')
    def test_unplug_interfaces(self, mock_update_nfd, mock_get_nfd):

        ndo_handler = self._initialize_ndo_handler()
        self.event = DummyEvent(dummy_data, 'ACTIVE')
        ndo_handler._prepare_device_data = mock.MagicMock(
            return_value=self.event.data)
        orig_event_data = copy.deepcopy(self.event.data)
        orig_event_data['status_description'] = (
            ndo_handler.status_map['ACTIVE'])
        orchestration_driver.unplug_network_function_device_interfaces = (
            mock.MagicMock(return_value=(True, [])))

        ndo_handler._controller.event_complete = mock.MagicMock(
            return_value=None)

        mock_get_nfd.return_value = {
            'id': 'device_id',
            'interfaces_in_use': 1,
            'reference_count': 1,
        }
        ndo_handler.unplug_interfaces(self.event)

        orig_event_data['interfaces_in_use'] -= len(orig_event_data['ports'])
        mock_update_nfd.assert_called_with(ndo_handler.db_session,
                                           orig_event_data['id'], mock.ANY)

        orig_event_data = copy.deepcopy(self.event.data)
        orig_event_data['status_description'] = (
            ndo_handler.status_map['ACTIVE'])
        orchestration_driver.unplug_network_function_device_interfaces = (
            mock.MagicMock(return_value=(False, [])))

        ndo_handler.unplug_interfaces(self.event)
        mock_update_nfd.assert_called_with(ndo_handler.db_session,
                                           orig_event_data['id'],
                                           {'interfaces_in_use': (
                                               orig_event_data[
                                                   'interfaces_in_use'])})

    """
    @mock.patch.object(nfpdb.NFPDbBase, 'delete_network_function_device')
    def test_device_delete(self, mock_delete_nfd):
        ndo_handler = self._initialize_ndo_handler()
        self.event = DummyEvent(101, 'ACTIVE', 1)
        ndo_handler._prepare_device_data = mock.MagicMock(
            return_value=self.event.data)
        orig_event_data = copy.deepcopy(self.event.data)
        orchestration_driver.delete_network_function_device = (
            mock.MagicMock(return_value=True))
        ndo_handler._create_event = mock.MagicMock(return_value=True)

        ndo_handler.delete_device(self.event)

        event_id = 'DEVICE_DELETED'
        orig_event_data['reference_count'] -= 1

        mock_delete_nfd.assert_called_with(ndo_handler.db_session,
                                           self.event.data['id'])
        ndo_handler._create_event.assert_called_with(event_id=event_id,
                                             event_data=orig_event_data)
    """

    def test_handle_device_create_error(self):
        ndo_handler = self._initialize_ndo_handler()
        event_id = status = 'DEVICE_CREATE_FAILED'
        self.event = DummyEvent(dummy_data, status, 1)
        orig_event_data = copy.deepcopy(self.event.data)
        orig_event_data['network_function_device_id'] = orig_event_data['id']
        ndo_handler._create_event = mock.MagicMock(return_value=True)

        ndo_handler.handle_device_create_error(self.event)
        ndo_handler._create_event.assert_called_with(
            event_id=event_id,
            event_data=orig_event_data)

    @mock.patch.object(nfpdb.NFPDbBase, 'update_network_function_device')
    def test_handle_device_not_up(self, mock_update_nfd):
        ndo_handler = self._initialize_ndo_handler()
        ndo_handler._controller.new_event = mock.MagicMock(
                    return_value=self.event)
        ndo_handler._controller.event_complete = mock.MagicMock(
                    return_value=None)
        status = 'ERROR'
        desc = 'Device not became ACTIVE'
        self.event = DummyEvent(dummy_data, status, 1)
        orig_event_data = copy.deepcopy(self.event.data)
        orig_event_data['status_description'] = desc
        orig_event_data.pop('interfaces_in_use', None)
        orig_event_data.pop('reference_count', None)
        ndo_handler._create_event = mock.MagicMock(return_value=True)

        ndo_handler.handle_device_not_up(self.event)
        orig_event_data['network_function_device_id'] = orig_event_data['id']
        mock_update_nfd.assert_called_with(ndo_handler.db_session,
                                           orig_event_data['id'],
                                           orig_event_data)

    @mock.patch.object(nfpdb.NFPDbBase, 'update_network_function_device')
    def test_handle_device_not_reachable(self, mock_update_nfd):
        ndo_handler = self._initialize_ndo_handler()
        status = 'ERROR'
        self.event = DummyEvent(dummy_data, status, 1)
        desc = 'Device not reachable, Health Check Failed'
        tmp_data = copy.deepcopy(self.event.data)
        device = self.event.data
        device = {'nfp_context': device,
                  'id': self.event.data['network_function_device']['id'],
                  'network_function_device_id': (
                      self.event.data['network_function_device']['id'])}
        device['nfp_context']['network_function_device']['reference_count'] = 0
        device['nfp_context']['network_function_device']['status'] = status
        device['nfp_context']['network_function_device'][
            'status_description'] = desc
        device['nfp_context']['network_function_device'][
            'reference_count'] += 1
        event_desc = Desc()
        device['nfp_context']['event_desc'] = event_desc.to_dict()
        device['nfp_context']['key'] = self.event.key
        ndo_handler._create_event = mock.MagicMock(return_value=True)
        ndo_handler._controller = mock.MagicMock(return_value=True)
        self.event.data = device
        ndo_handler.handle_device_not_reachable(self.event)
        device['nfp_context']['network_function_device'].pop(
            'reference_count', None)
        device['nfp_context']['network_function_device'].pop(
            'interfaces_in_use', None)
        mock_update_nfd.assert_called_with(ndo_handler.db_session,
                                           device['nfp_context'][
                                               'network_function_device'][
                                                   'id'],
                                           device)
        ndo_handler._controller.reset_mock()

        self.event.data = tmp_data

    @mock.patch.object(nfpdb.NFPDbBase, '_get_network_function_device')
    @mock.patch.object(nfpdb.NFPDbBase, 'update_network_function_device')
    def test_handle_device_config_failed(self, mock_update_nfd, mock_get_nfd):
        ndo_handler = self._initialize_ndo_handler()
        status = 'ERROR'
        self.event = DummyEvent(dummy_data, status, 1)
        desc = 'Configuring Device Failed.'
        tmp_data = copy.deepcopy(self.event.data)
        device = self.event.data
        device = {'nfp_context': device,
                  'id': self.event.data['network_function_device']['id']}
        device['nfp_context']['network_function_device']['reference_count'] = 0
        device['nfp_context']['network_function_device']['status'] = status
        device['nfp_context']['network_function_device'][
            'status_description'] = desc
        reference_count = device['nfp_context']['network_function_device'][
            'reference_count'] + 1
        event_desc = Desc()
        device['nfp_context']['event_desc'] = event_desc.to_dict()
        device['nfp_context']['key'] = self.event.key
        device['nfp_context']['binding_key'] = self.event.binding_key
        ndo_handler._create_event = mock.MagicMock(return_value=True)
        ndo_handler._controller = mock.MagicMock(return_value=True)
        ndo_handler.nsf_db.get_network_function_device = (
            mock.MagicMock(return_value={'reference_count': (
                reference_count - 1)}))
        self.event.data = device
        mock_get_nfd.return_value = {
            'id': 'device_id',
            'reference_count': 0
        }
        ndo_handler.handle_device_config_failed(self.event)
        mock_update_nfd.assert_called_with(ndo_handler.db_session,
                                           device['nfp_context'][
                                               'network_function_device'][
                                                   'id'],
                                           {'status': 'ERROR',
                                            'status_description':
                                                'Configuring Device Failed.',
                                            'id': 'vm-id'})
        ndo_handler._controller.reset_mock()

        self.event.data = tmp_data


def main():
    unittest2.main()


if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import multiprocessing
import uuid as pyuuid

from gbpservice.nfp.core import common as nfp_common
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import module as nfp_api
from gbpservice.nfp.core import sequencer as nfp_seq

LOG = nfp_logging.getLogger(__name__)
identify = nfp_common.identify

"""Event Types """
SCHEDULE_EVENT = 'schedule_event'
POLL_EVENT = 'poll_event'
STASH_EVENT = 'stash_event'

"""Event Flag """
EVENT_NEW = 'new_event'
EVENT_COMPLETE = 'event_done'
EVENT_ACK = 'event_ack'
POLL_EVENT_STOP = 'poll_event_stop'

EVENT_DEFAULT_LIFETIME = 600

"""Sequencer status. """
SequencerEmpty = nfp_seq.SequencerEmpty
SequencerBusy = nfp_seq.SequencerBusy

deque = collections.deque

"""Defines poll descriptor of an event.

    Holds all of the polling information of an
    event.
"""


class PollDesc(object):

    def __init__(self, **kwargs):
        # Spacing of the event, event will timeout @this spacing.
        self.spacing = kwargs.get('spacing')
        # Max times event can be polled, is autocancelled after.
        self.max_times = kwargs.get('max_times')
        # Reference to original event, UUID.
        self.ref = kwargs.get('ref')


"""Defines the descriptor of an event.

    Holds the metadata for an event. Useful
    for event processing. Not exposed to nfp modules.
"""


class EventDesc(object):

    def __init__(self, **kwargs):
        # Unique id of the event, use what user passed or
        # generate a new unique id.
        uuid = kwargs.get('key', pyuuid.uuid4())
        id = kwargs.get('id', '')

        self.uuid = str(uuid) + ':' + id

        # see 'Event Types'
        self.type = kwargs.get('type')
        # see 'Event Flag'
        self.flag = kwargs.get('flag')
        # PID of worker which is handling this event
        self.worker = kwargs.get('worker')
        # Polling descriptor of event
        self.poll_desc = kwargs.get('poll_desc')
        # Target module to which this event must be delivered
        self.target = None
        # ID of graph of which this event is part of
        self.graph = None
        # Type of path to which this event belongs CREATE/UPDATE/DELETE
        self.path_type = kwargs.get('path_type')
        # Unique key for the path
        self.path_key = kwargs.get('path_key')
        # Marks whether an event was acked or not
        self.acked = False

    def from_desc(self, desc):
        self.type = desc.type
        self.flag = desc.flag
        self.worker = desc.worker
        self.poll_desc = desc.poll_desc
        self.path_type = desc.path_type
        self.path_key = desc.path_key

    def to_dict(self):
        return {'uuid': self.uuid,
                'type': self.type,
                'flag': self.flag,
                'worker': self.worker,
                'poll_desc': self.poll_desc,
                'path_type': self.path_type,
                'path_key': self.path_key
                }


"""Defines the event structure.

    Nfp modules need to create object of the class
    to create an event.
"""


class Event(object):

    def __init__(self, **kwargs):
        # ID of event as passed by module
        self.id = kwargs.get('id')
        # Data blob
        self.data = kwargs.get('data')
        # Whether to sequence this event w.r.t
        # other related events.
        self.sequence = kwargs.get('serialize', False)
        # Unique key to be associated with the event
        self.key = kwargs.get('key')
        # Binding key to define relation between
        # different events.
        self.binding_key = kwargs.get('binding_key')
        # Handler of the event.
        self.handler = kwargs.get('handler')
        # Lifetime of the event in seconds.
        self.lifetime = kwargs.get('lifetime', -1)
        # Identifies whether event.data is zipped
        self.zipped = False
        # Log metadata context
        self.context = kwargs.get('context', {})
        # Prepare the base descriptor
        desc = kwargs.get('desc_dict')
        if desc:
            desc['key'] = self.key
            desc['id'] = self.id
            desc = EventDesc(**desc)
        elif self.key:
            desc = EventDesc(**{'key': self.key,
                                'id': self.id})
        else:
            desc = EventDesc(**{'id': self.id})
        self.desc = desc

        self.result = None

        cond = self.sequence is True and self.binding_key is None
        assert not cond

    def identify(self):
        if hasattr(self, 'desc'):
            return "uuid=%s,id=%s,type=%s,flag=%s" % (
                self.desc.uuid, self.id, self.desc.type, self.desc.flag)
        return "id=%s" % (self.id)


"""Table of event handler's.

    Maintains cache of every module's event handlers.
    Also, maintains the polling against event_id
    which are provided as decorators.
"""


class NfpEventHandlers(object):

    def __init__(self):
        # {'event.id': [(event_handler, poll_handler, spacing)]
        self._event_desc_table = {}

    def _log_meta(self, event_id, event_handler=None):
        if event_handler:
            return "(event_id - %s) - (event_handler - %s)" % (
                event_id, identify(event_handler))
        else:
            return "(event_id - %s) - (event_handler - None)" % (event_id)

    def register(self, event_id, event_handler, module, priority=0):
        """Registers a handler for event_id.

            Also fetches the decorated poll handlers if any
            for the event and caches it.
        """
        if not isinstance(event_handler, nfp_api.NfpEventHandler):
            message = "%s - Handler is not instance of NfpEventHandler" % (
                self._log_meta(event_id, event_handler))
            LOG.error(message)
            return
        try:
            poll_desc_table = event_handler.get_poll_desc_table()
            poll_handler = poll_desc_table[event_id]
            spacing = poll_handler._spacing
        except KeyError:
            # Default the poll handler and spacing values
            poll_handler = event_handler.handle_poll_event
            spacing = 0

        try:
            try:
                self._event_desc_table[event_id]['modules'][module].append(
                    (event_handler, poll_handler, spacing, module))
            except KeyError:
                self._event_desc_table[event_id]['modules'][module] = [
                    (event_handler, poll_handler, spacing, module)]
            try:
                self._event_desc_table[event_id]['priority'][priority].append(
                    (event_handler, poll_handler, spacing, module))
            except KeyError:
                self._event_desc_table[event_id]['priority'][priority] = [
                    (event_handler, poll_handler, spacing, module)]
        except KeyError:
            self._event_desc_table[event_id] = {'modules': {}, 'priority': {}}
            self._event_desc_table[event_id]['modules'][module] = [
                (event_handler, poll_handler, spacing, module)]
            self._event_desc_table[event_id]['priority'][priority] = [
                (event_handler, poll_handler, spacing, module)]
        message = "%s - Registered handler" % (
            self._log_meta(event_id, event_handler))
        LOG.debug(message)

    def get_event_handler(self, event_id, module=None):
        """Get the handler for the event_id. """
        eh = None
        rmodule = None
        try:
            if module:
                eh = self._event_desc_table[event_id]['modules'][module][0][0]
                rmodule = (
                    self._event_desc_table[event_id]['modules'][module][0][3])
            else:
                priorities = (
                    self._event_desc_table[event_id]['priority'].keys())
                priority = max(priorities)
                eh = (
                    self._event_desc_table[
                        event_id]['priority'][priority][0][0])
                rmodule = (
                    self._event_desc_table[
                        event_id]['priority'][priority][0][3])
        finally:
            message = "%s - Returning event handler" % (
                self._log_meta(event_id, eh))
            LOG.debug(message)
            return eh, rmodule

    def get_poll_handler(self, event_id, module=None):
        """Get the poll handler for event_id. """
        ph, spacing = None, None
        try:
            if module:
                ph = self._event_desc_table[event_id]['modules'][module][0][1]
                spacing = self._event_desc_table[
                    event_id]['modules'][module][0][2]
            else:
                priorities = (
                    self._event_desc_table[event_id]['priority'].keys())
                priority = max(priorities)
                ph = (
                    self._event_desc_table[
                        event_id]['priority'][priority][0][1])
                spacing = self._event_desc_table[
                    event_id]['priority'][priority][0][2]
        finally:
            message = "%s - Returning poll handler" % (
                self._log_meta(event_id, ph))
            LOG.debug(message)
            return ph, spacing

    def get_poll_spacing(self, event_id):
        """Return the spacing for event_id. """
        spacing = 0
        try:
            spacing = self._event_desc_table[event_id][0][2]
        finally:
            message = "%s - Poll spacing %d" % (
                self._log_meta(event_id), spacing)
            LOG.debug(message)
            return spacing


"""Manages the lifecycle of event of a process.

    Each process (worker/distributor) is associated
    with a event manager. Event manager pulls events
    from the pipe, caches it, sequences & dispatches
    the events.
"""


class NfpEventManager(object):

    def __init__(self, conf, controller, sequencer, pipe=None, pid=-1):
        self._conf = conf
        self._controller = controller
        # PID of process to which this event manager is associated
        self._pid = pid
        # Duplex pipe to read & write events
        self._pipe = pipe
        # Cache of UUIDs of events which are dispatched to
        # the worker which is handled by this em.
        self._cache = deque()
        # Load on this event manager - num of events pending to be completed
        self._load = 0

    def _log_meta(self, event=None):
        if event:
            return "(event - %s) - (event_manager - %d)" % (
                event.identify(), self._pid)
        else:
            return "(event_manager - %d" % (self._pid)

    def _wait_for_events(self, pipe, timeout=0.01):
        """Wait & pull event from the pipe.

            Wait till timeout for the first event and then
            pull as many as available.
            Returns: Events[] pulled from pipe.
        """
        events = []
        try:
            ret = pipe.poll(timeout)
            if ret:
                event = self._controller.pipe_recv(pipe)
                if event:
                    events.append(event)
        except multiprocessing.TimeoutError as err:
            message = "%s" % (err)
            LOG.exception(message)
        return events

    def init_from_event_manager(self, em):
        """Initialize from existing event manager.

            Invoked when an event manager has to take over
            existing event manager.

            Whole cache is replaced and events are replayed.
            This is used in case where a worker dies, dead
            workers event manager is assigned to new worker.
        """
        # Replay all the events from cache.
        self._cache = em._cache

    def get_pending_events(self):
        return list(self._cache)

    def get_load(self):
        """Return current load on the manager."""
        return self._load

    def pop_event(self, event):
        """Pop the passed event from cache.

            Is called when an event is complete/cancelled.
            If the event was sequenced, then sequencer is
            released to schedule next event.

            Removes event from cache.
        """
        message = "%s - pop event" % (self._log_meta(event))
        LOG.debug(message)
        try:
            self._cache.remove(event.desc.uuid)
            self._load -= 1
        except ValueError as verr:
            verr = verr
            message = "%s - event not in cache" % (
                self._log_meta(event))
            LOG.debug(message)

    def dispatch_event(self, event, event_type=None,
                       inc_load=True, cache=True):
        """Dispatch event to the worker.

            Sends the event to worker through pipe.
            Increments load if event_type is SCHEDULED event,
            poll_event does not contribute to load.
        """
        message = "%s - Dispatching to worker %d" % (
            self._log_meta(event), self._pid)
        LOG.debug(message)
        # Update the worker information in the event.
        event.desc.worker = self._pid
        # Update the event with passed type
        if event_type:
            event.desc.type = event_type
        # Send to the worker
        self._controller.pipe_send(self._pipe, event)

        self._load = (self._load + 1) if inc_load else self._load
        # Add to the cache
        if cache:
            self._cache.append(event.desc.uuid)

    def event_watcher(self, timeout=0.01):
        """Watch for events. """
        return self._wait_for_events(self._pipe, timeout=timeout)
